import subprocess
import json
from typing import List, Any


class OPLookupError(LookupError):
    pass


class TooManyEntriesOPLookupError(OPLookupError):
    pass


class NoEntriesOPLookupError(OPLookupError):
    pass


class NoFieldValueOPLookupError(OPLookupError):
    pass


def op_smart_lookup(env_var_name: str) -> str:
    fields = _op_fields_to_try(env_var_name)
    for field in fields:
        try:
            return op_lookup(env_var_name, field_name=field)
        except NoFieldValueOPLookupError:
            pass
    raise NoFieldValueOPLookupError('1Passsword entry with tag '
                                    f'{env_var_name} has no value for '
                                    'the fields tried: '
                                    f'{", ".join(fields)}.  Please populate '
                                    'one of these fields in 1Password.')


def _op_list_items(env_var_name: str) -> List[Any]:
    list_command = ['op', 'list', 'items', '--tags', env_var_name]
    list_output = subprocess.check_output(list_command)
    list_output_data = json.loads(list_output)
    assert isinstance(list_output_data, list)
    if len(list_output_data) == 0:
        raise NoEntriesOPLookupError(f"No 1Password entries with tag {env_var_name} found")
    if len(list_output_data) > 1:
        raise TooManyEntriesOPLookupError("Too many 1Password entries "
                                          f"with tag {env_var_name} found")
    return list_output_data


def _op_get_item_fields(list_output: bytes, field_name: str = 'password') -> bytes:
    get_command = ['op', 'get', 'item', '-', '--fields', field_name]
    return subprocess.check_output(get_command, input=list_output)


def op_lookup(env_var_name: str, field_name: str = 'password') -> str:
    # https://stackoverflow.com/questions/13332268/how-to-use-subprocess-command-with-pipes
    list_output_data = _op_list_items(env_var_name)
    list_output = json.dumps(list_output_data).encode('utf-8')
    get_output = _op_get_item_fields(list_output, field_name)
    get_output_str = get_output.decode('utf-8').rstrip('\n')
    if get_output_str == '':
        raise NoFieldValueOPLookupError('1Passsword entry with tag '
                                        f'{env_var_name} has no value for field {field_name}')
    return get_output_str


def _name_inferred_fields(env_var_name: str) -> List[str]:
    components = env_var_name.split('_')
    if len(components) <= 1:
        return []
    conversions = {
        'user': 'username'
    }
    raw_value = components[-1].lower()
    return [conversions.get(raw_value, raw_value)]


def _op_fields_to_try(env_var_name: str) -> List[str]:
    return _name_inferred_fields(env_var_name) + ['password']



#!/usr/bin/env python

"""Tests for `op_env` package."""

import os
import pytest
import subprocess
from unittest.mock import patch, call

from op_env._cli import parse_argv, process_args
from op_env.op import (
    op_lookup,
    op_smart_lookup,
    _op_fields_to_try,
    TooManyEntriesOPLookupError,
    NoEntriesOPLookupError,
    NoFieldValueOPLookupError,
)


def test_process_args_runs_simple_command_with_simple_env():
    with patch('op_env._cli.subprocess') as mock_subprocess,\
         patch('op_env._cli.op_smart_lookup') as mock_op_lookup:
        command = ['env']
        args = {'operation': 'run', 'command': command,
                'environment': ['a']}
        process_args(args)
        mock_op_lookup.assert_called_with('a')
        mock_subprocess.check_call.assert_called_with(command,
                                                      env={'a': mock_op_lookup.return_value})


def test_process_args_runs_simple_command():
    with patch('op_env._cli.subprocess') as mock_subprocess:
        command = ['env']
        args = {'operation': 'run', 'command': command,
                'environment': []}
        process_args(args)
        mock_subprocess.check_call.assert_called_with(command, env={})


def test_process_args_rejects_non_run():
    with patch('op_env.op.subprocess'):  # for safety
        with pytest.raises(ValueError):
            args = {'operation': 'definitely-not-run'}
            process_args(args)


def test_fields_to_try_conversion_username():
    with patch('op_env.op.subprocess'):  # for safety
        out = _op_fields_to_try('ABC_USER')
        assert out == ['username', 'password']


def test_fields_to_try_multiple_words():
    with patch('op_env.op.subprocess'):  # for safety
        out = _op_fields_to_try('ABC_FLOOGLE')
        assert out == ['floogle', 'password']


def test_fields_to_try_simple():
    with patch('op_env.op.subprocess'):  # for safety
        out = _op_fields_to_try('ABC')
        assert out == ['password']


def test_op_lookup_no_field_value():
    with patch('op_env.op.subprocess') as mock_subprocess:
        list_output = b"[{}]"
        get_output = b"\n"
        mock_subprocess.check_output.side_effect = [
            list_output,
            get_output,
        ]
        with pytest.raises(NoFieldValueOPLookupError,
                           match=('1Passsword entry with tag '
                                  'ANY_TEST_VALUE has no value for field abc')):
            op_lookup('ANY_TEST_VALUE', field_name='abc')
        mock_subprocess.check_output.\
            assert_has_calls([call(['op', 'list', 'items', '--tags', 'ANY_TEST_VALUE']),
                              call(['op', 'get', 'item', '-', '--fields', 'abc'],
                                   input=list_output)])


def test_op_lookup_too_few_entries():
    with patch('op_env.op.subprocess') as mock_subprocess:
        list_output = b"[]"
        mock_subprocess.check_output.return_value = list_output
        with pytest.raises(NoEntriesOPLookupError,
                           match='No 1Password entries with tag ANY_TEST_VALUE found'):
            op_lookup('ANY_TEST_VALUE', field_name='abc')
        mock_subprocess.check_output.\
            assert_called_with(['op', 'list', 'items', '--tags', 'ANY_TEST_VALUE'])


def test_op_lookup_too_many_entries():
    with patch('op_env.op.subprocess') as mock_subprocess:
        list_output = b"[{}, {}]"
        mock_subprocess.check_output.return_value = list_output
        with pytest.raises(TooManyEntriesOPLookupError,
                           match='Too many 1Password entries with tag ANY_TEST_VALUE'):
            op_lookup('ANY_TEST_VALUE', field_name='abc')
        mock_subprocess.check_output.\
            assert_called_with(['op', 'list', 'items', '--tags', 'ANY_TEST_VALUE'])


def test_op_lookup_specific_field():
    with patch('op_env.op.subprocess') as mock_subprocess:
        list_output = b"[{}]"
        get_output = b"get_results\n"
        mock_subprocess.check_output.side_effect = [
            list_output,
            get_output,
        ]
        out = op_lookup('ANY_TEST_VALUE', field_name='abc')
        mock_subprocess.check_output.\
            assert_has_calls([call(['op', 'list', 'items', '--tags', 'ANY_TEST_VALUE']),
                              call(['op', 'get', 'item', '-', '--fields', 'abc'],
                                   input=list_output)])
        assert out == "get_results"


def test_op_smart_lookup_multiple_fields():
    with patch('op_env.op.op_lookup') as mock_op_lookup,\
         patch('op_env.op._op_fields_to_try') as mock_op_fields_to_try:
        mock_op_fields_to_try.return_value = ['floogle', 'blah']
        mock_op_lookup.return_value = 'result value'
        ret = op_smart_lookup('ENVVARNAME')
        mock_op_fields_to_try.assert_called_with('ENVVARNAME')
        mock_op_lookup.assert_called_with('ENVVARNAME', field_name='floogle')
        assert ret == 'result value'


def test_op_smart_lookup_multiple_fields_all_errors():
    with patch('op_env.op.op_lookup') as mock_op_lookup,\
         patch('op_env.op._op_fields_to_try') as mock_op_fields_to_try:
        mock_op_fields_to_try.return_value = ['floogle', 'blah']
        mock_op_lookup.side_effect = [NoFieldValueOPLookupError,
                                      NoFieldValueOPLookupError]
        with pytest.raises(NoFieldValueOPLookupError,
                           match=('1Passsword entry with tag '
                                  'ENVVARNAME has no value for '
                                  'the fields tried: '
                                  "floogle, blah.  Please populate "
                                  'one of these fields in 1Password.')):
            op_smart_lookup('ENVVARNAME')
        mock_op_fields_to_try.assert_called_with('ENVVARNAME')
        mock_op_lookup.assert_has_calls([call('ENVVARNAME', field_name='floogle'),
                                         call('ENVVARNAME', field_name='blah')])


def test_op_smart_lookup_single_field_with_error():
    with patch('op_env.op.op_lookup') as mock_op_lookup,\
         patch('op_env.op._op_fields_to_try') as mock_op_fields_to_try:
        mock_op_fields_to_try.return_value = ['floogle']
        mock_op_lookup.side_effect = NoFieldValueOPLookupError
        with pytest.raises(NoFieldValueOPLookupError):
            op_smart_lookup('ENVVARNAME')
        mock_op_fields_to_try.assert_called_with('ENVVARNAME')
        mock_op_lookup.assert_called_with('ENVVARNAME', field_name='floogle')


def test_op_smart_lookup_multiple_fields_chooses_second():
    with patch('op_env.op.op_lookup') as mock_op_lookup,\
         patch('op_env.op._op_fields_to_try') as mock_op_fields_to_try:
        mock_op_fields_to_try.return_value = ['floogle', 'blah']
        mock_op_lookup.side_effect = [NoFieldValueOPLookupError,
                                      'result value']
        ret = op_smart_lookup('ENVVARNAME')
        mock_op_fields_to_try.assert_called_with('ENVVARNAME')
        mock_op_lookup.assert_has_calls([call('ENVVARNAME', field_name='floogle'),
                                         call('ENVVARNAME', field_name='blah')])
        assert ret == 'result value'


def test_op_smart_lookup_chooses_first():
    with patch('op_env.op.op_lookup') as mock_op_lookup,\
         patch('op_env.op._op_fields_to_try') as mock_op_fields_to_try:
        mock_op_fields_to_try.return_value = ['floogle']
        ret = op_smart_lookup('ENVVARNAME')
        mock_op_fields_to_try.assert_called_with('ENVVARNAME')
        mock_op_lookup.assert_called_with('ENVVARNAME', field_name='floogle')
        assert ret == mock_op_lookup.return_value


def test_parse_args_run_command_with_long_env_variables():
    argv = ['op-env', 'run', '-e', 'DUMMY', '--environment', 'DUMMY2', 'mycmd']
    args = parse_argv(argv)
    assert args == {'command': ['mycmd'],
                    'environment': ['DUMMY', 'DUMMY2'],
                    'operation': 'run'}


def test_parse_args_run_command_no_env_variables():
    argv = ['op-env', 'run', 'mycmd']
    args = parse_argv(argv)
    assert args == {'command': ['mycmd'],
                    'environment': [],
                    'operation': 'run'}


def test_parse_args_run_command_with_multiple_variables():
    argv = ['op-env', 'run', '-e', 'DUMMY', '-e', 'DUMMY2', 'mycmd']
    args = parse_argv(argv)
    assert args == {'command': ['mycmd'],
                    'environment': ['DUMMY', 'DUMMY2'],
                    'operation': 'run'}


def test_parse_args_run_command_with_arguments():
    argv = ['op-env', 'run', '-e', 'DUMMY', 'mycmd', '1', '2', '3']
    args = parse_argv(argv)
    assert args == {'command': ['mycmd', '1', '2', '3'],
                    'environment': ['DUMMY'],
                    'operation': 'run'}


def test_parse_args_run_simple():
    argv = ['op-env', 'run', '-e', 'DUMMY', 'mycmd']
    args = parse_argv(argv)
    assert args == {'command': ['mycmd'], 'environment': ['DUMMY'], 'operation': 'run'}


@pytest.mark.skip(reason="need to mock op binary in test PATH")
def test_cli_run():
    argv = ['op-env', 'run', '-e', 'DUMMY', 'env']
    expected_envvar = 'DUMMY=dummyvalue'
    actual_output = subprocess.check_output(argv).decode('utf-8')
    assert expected_envvar in actual_output


def test_cli_help_run():
    expected_help = """usage: op-env run [-h] [--environment ENVVAR] command [command ...]

positional arguments:
  command               Command to run with the environment set from 1Password

optional arguments:
  -h, --help            show this help message and exit
  --environment ENVVAR, -e ENVVAR
                        environment variable name to set, based on item with same tag in 1Password
"""
    request_long_lines = {'COLUMNS': '999', 'LINES': '25'}
    env = {}
    env.update(os.environ)
    env.update(request_long_lines)

    # older python versions show arguments like this:
    actual_help = subprocess.check_output(['op-env', 'run', '--help'], env=env).decode('utf-8')
    assert actual_help == expected_help


def test_cli_no_args():
    expected_help = """usage: op-env [-h] {run} ...
op-env: error: the following arguments are required: operation
"""
    request_long_lines = {'COLUMNS': '999', 'LINES': '25'}
    env = {}
    env.update(os.environ)
    env.update(request_long_lines)

    # older python versions show arguments like this:
    completed_process = subprocess.run(['op-env'], env=env,
                                       stderr=subprocess.PIPE)
    actual_help = completed_process.stderr.decode('utf-8')
    assert actual_help == expected_help
    assert completed_process.returncode == 2


def test_cli_help():
    expected_help = """usage: op-env [-h] {run} ...

positional arguments:
  {run}       Run the specified command with the given environment variables

optional arguments:
  -h, --help  show this help message and exit
"""
    request_long_lines = {'COLUMNS': '999', 'LINES': '25'}
    env = {}
    env.update(os.environ)
    env.update(request_long_lines)

    # older python versions show arguments like this:
    actual_help = subprocess.check_output(['op-env', '--help'],
                                          env=env).decode('utf-8')
    assert actual_help == expected_help



"""Console script for op_env."""
import argparse
import sys
from typing import List, Dict
import subprocess
from .op import op_smart_lookup


def parse_argv(argv: List[str]) -> Dict[str, str]:
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(help='Run the specified command '
                                       'with the given environment variables',
                                       dest='operation')
    subparsers.required = True
    run_parser = subparsers.add_parser('run')
    run_parser.add_argument('--environment', '-e',
                            metavar='ENVVAR',
                            action='append',
                            default=[],
                            help='environment variable name to set, '
                            'based on item with same tag in 1Password')
    run_parser.add_argument('command',
                            nargs='+',
                            help='Command to run with the environment set from 1Password')
    return vars(parser.parse_args(argv[1:]))


def process_args(args: Dict[str, str]) -> int:
    if args['operation'] == 'run':
        env: Dict[str, str] = {
            envvar: op_smart_lookup(envvar)
            for envvar in args['environment']
        }
        subprocess.check_call(args['command'], env=env)
        return 0
    else:
        raise ValueError(f"Unknown operation: {args['operation']}")


def main(argv: List[str] = sys.argv) -> int:
    """Console script for op_env."""
    args = parse_argv(argv)
    return process_args(args)


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages
import distutils
from distutils.cmd import Command
import distutils.command.clean
from distutils.dir_util import remove_tree
import subprocess
import os
from typing import List, Tuple, Optional

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements: List[str] = []

setup_requirements: List[str] = ['pytest-runner']

test_requirements: List[str] = ['pytest>=3']


class MypyCleanCommand(Command):
    """Regular clean plus mypy cache"""

    description = 'Run mypy on source code'
    user_options: List[Tuple[str, Optional[str], str]] = []

    def initialize_options(self) -> None:
        pass

    def finalize_options(self) -> None:
        pass

    def run(self) -> None:
        if os.path.exists('.mypy_cache'):
            remove_tree('.mypy_cache')


class MypyCommand(Command):
    description = 'Run mypy on source code'
    user_options: List[Tuple[str, Optional[str], str]] = []

    def initialize_options(self) -> None:
        pass

    def finalize_options(self) -> None:
        pass

    def run(self) -> None:
        """Run command."""
        command = ['mypy', '--html-report', 'types/coverage', '.']
        self.announce(
            'Running command: %s' % str(command),
            level=distutils.log.INFO)  # type: ignore
        subprocess.check_call(command)


class QualityCommand(Command):
    quality_target: Optional[str]

    description = 'Run quality tools on source code'
    user_options: List[Tuple[str, Optional[str], str]] = []

    def initialize_options(self) -> None:
        pass

    def finalize_options(self) -> None:
        pass

    def run(self) -> None:
        """Run command."""
        command = ['overcommit', '--run']
        self.announce(
            'Running command: %s' % str(command),
            level=distutils.log.INFO)  # type: ignore
        subprocess.check_call(command)


setup(
    author="Vince Broz",
    author_email='vince@broz.cc',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description="op-env allows you to use OnePassword entries as environment variable-style secrets",  # noqa: E501
    entry_points={
        'console_scripts': [
            'op-env=op_env._cli:main',
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='op_env',
    name='op_env',
    packages=find_packages(include=['op_env',
                                    'op_env.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/apiology/op_env',
    version='0.1.0',
    zip_safe=False,
    cmdclass={
        'quality': QualityCommand,
        'typesclean': MypyCleanCommand,
        'types': MypyCommand,
    },
)
